#include <iostream>
#include <QStringList>

#include "treeitem.h"

#include "branchobj.h"
#include "branchitem.h"
#include "vymmodel.h"

using namespace std;

extern FlagRow* standardFlagsMaster;

TreeItem::TreeItem()
{
	//cout << "Constr. TI\n";
	init();
	itemData.clear();
	rootItem=this;
	parentItem=NULL;
}

TreeItem::TreeItem(const QList<QVariant> &data, TreeItem *parent)
{
	//cout << "Constructor TreeItem this="<<this<<"  parent="<<parent<<endl;
	init();
    parentItem = parent;
    itemData = data;
	
	rootItem=this;
	if (parentItem )
		rootItem=parentItem->rootItem;
}

TreeItem::~TreeItem()
{
	//cout << "Destructor TreeItem "<<getHeadingStd()<<endl;
	TreeItem *ti;
	while (!childItems.isEmpty())
	{
		ti=childItems.takeFirst();
		switch (ti->getType() )
		{
			case TreeItem::MapCenter:
				delete (BranchItem*)ti;
				break;
			case TreeItem::Branch:
				delete (BranchItem*)ti;
				break;
			default:
				delete ti;
				break;
		}
	}	
}


void TreeItem::init()
{
	model=NULL;

    parentItem = NULL;
    itemData.clear();
	QList<QVariant> cData;
	itemData<< "VM:addMapCenter" << "undef"<<"undef";


	branchOffset=0;
	branchCounter=0;
	lastSelectedBranchNum=-1;

	imageOffset=0;
	imageCounter=0;

	note.setNote(""); 
	// note.setFontHint (textEditor->getFontHintDefault() );	//FIXME-2
	// isNoteInEditor=false;

	hidden=false;
	hideExport=false;

	// Reset ID
	objID="";

	standardFlags.setMasterRow (standardFlagsMaster);
}


/*
void TreeItem::copy (OrnamentedObj* other)	//FIXME-2	probably need deep copy of branches and data!

{
	note.copy (other->note);
	model=other->model;
	hideExport=officially 
	other->hideExport;
}
*/

void TreeItem::setModel (VymModel *m)
{
	model=m;
}

VymModel* TreeItem::getModel ()
{
	return model;
}

int TreeItem::getRowNumAppend (TreeItem *item)
{	
	switch (item->type)
	{
		case MapCenter: return branchOffset + branchCounter;
		case Branch: return branchOffset + branchCounter;
		case Image: return imageOffset + imageCounter;
		default: return -1;
	}
}

void TreeItem::appendChild(TreeItem *item)
{
	item->parentItem=this;
	item->rootItem=rootItem;
	item->setModel (model);

	if (item->type == Image)
	{
		childItems.insert (imageCounter,item);
		// images are on top of list
		if (imageCounter==0)
			imageOffset=0;
		imageCounter++;
		branchOffset++;
	}

	if (item->isBranchLikeType())
	{
		// branches are on bottom of list
		childItems.append(item);
		if (branchCounter==0)
			branchOffset=childItems.count()-1;
		branchCounter++;
	}
}

void TreeItem::removeChild(int row)
{
    if (row<0 || row > childItems.size()-1)
		qWarning ("TreeItem::removeChild tried to remove non existing item?!\n");
	else
	{
		if (childItems.at(row)->isBranchLikeType())
			branchCounter--;
		if (childItems.at(row)->type==Image)
		{
			imageCounter--;
			if (branchOffset>0) 
				branchOffset--;
		}	
		childItems.removeAt (row);
	}
}

void TreeItem::removeChildBranches()
{
	while ( childItems.size()>0)
	{
		branchCounter--;
		childItems.removeAt (0);
	}
}

TreeItem *TreeItem::child(int row)
{
    return childItems.value(row);
}

int TreeItem::childCount() const
{
    return childItems.count();
}

int TreeItem::childNumber() const
{
    if (parentItem)
        return parentItem->childItems.indexOf(const_cast<TreeItem*>(this));

    return 0;
}

int TreeItem::columnCount() const
{
    return itemData.count();
}

int TreeItem::branchCount() const
{
    return branchCounter;
}

int TreeItem::imageCount() const
{
    return imageCounter; 
}

int TreeItem::xlinkCount() const // FIXME-2 check if xlinks are stored in a different way (global to model?)
{
	int xlinkCounter=0;
    return xlinkCounter; // FIXME-1 xlinkCounter needs to be calculated...
}

int TreeItem::row() const
{
    if (parentItem)
        return parentItem->childItems.indexOf(const_cast<TreeItem*>(this));

    return 0;
}

int TreeItem::column() const
{
    return 0;
}

int TreeItem::depth() 
{
	int d=-2;
	TreeItem *ti=this;
	while (ti!=NULL)
	{
		ti=ti->parent();
		d++;
	}
	return d;
}

TreeItem *TreeItem::parent()
{
    return parentItem;
}

int TreeItem::childNum()
{
	return parentItem->childItems.indexOf (this);
}

int TreeItem::num()
{
	switch (type)
	{
		case MapCenter: return parentItem->childItems.indexOf (this) - parentItem->branchOffset;
		case Branch: return parentItem->childItems.indexOf (this) - parentItem->branchOffset;
		case Image: return parentItem->childItems.indexOf (this) - parentItem->imageOffset;
		default: return -1;
	}
}

int TreeItem::num (TreeItem *item)
{
	if (!item) return -1;
	if (!childItems.contains(item)) return -1;
	switch (item->getType())
	{
		case MapCenter: return childItems.indexOf (item) - branchOffset;
		case Branch: return childItems.indexOf (item) - branchOffset;
		case Image: return parentItem->childItems.indexOf (item) - imageOffset;
		default: return -1;
	}
}
void TreeItem::setType(const Type t)
{
	type=t;
	itemData[1]=getTypeName();
}

TreeItem::Type TreeItem::getType()
{
	return type;
}

bool TreeItem::isBranchLikeType() const
{
	if (type==Branch ||type==MapCenter) return true;
	return false;
}

QString TreeItem::getTypeName()
{
	switch (type)
	{
		case Undefined: return QString ("Undefined");
		case MapCenter: return QString ("MapCenter");
		case Branch: return QString ("Branch");
		case Image: return QString ("Image");
		default: return QString ("TreeItem::getTypeName no typename defined?!");
	}
}


QVariant TreeItem::data(int column) const
{
    return itemData.value(column);
}

void TreeItem::setHeading (const QString s)
{
	itemData[0]=s;
}

QString TreeItem::getHeading () const
{
	return itemData[0].toString();
}

std::string TreeItem::getHeadingStd () const
{
	return itemData[0].toString().toStdString();
}

void TreeItem::setHeadingColor (QColor color)
{
	headingColor=color;
}

QColor TreeItem::getHeadingColor ()
{
	return headingColor;
}

void TreeItem::setURL (const QString &u)
{
	url=u;
	if (!url.isEmpty())
		systemFlags.activate ("system-url");
	else
		systemFlags.deactivate ("system-url");
}

QString TreeItem::getURL ()
{
	return url;
}

void TreeItem::setVymLink (const QString &vl)
{
	if (!vl.isEmpty())
	{
		// We need the relative (from loading) 
		// or absolute path (from User event)
		// and build the absolute path.
		// Note: If we have relative, use path of
		// current map to build absolute path
		QDir d(vl);
		if (!d.path().startsWith ("/"))
		{
			QString p=model->getDestPath();
			int i=p.findRev("/",-1);
			d.setPath(p.left(i)+"/"+vl);
			d.convertToAbs();
		}
		vymLink=d.path();
		systemFlags.activate("system-vymLink");
	}	
	else	
	{
		systemFlags.deactivate("system-vymLink");
		vymLink.clear();
	}	
}

QString TreeItem::getVymLink ()
{
	return vymLink;
}

void TreeItem::setNote(const QString &s)
{
	NoteObj n;
	n.setNote(s);
	setNoteObj (n,false);
}

void TreeItem::clearNote()
{
	note.clear();
	systemFlags.deactivate ("system-note");
}

void TreeItem::setNoteObj(const NoteObj &n, bool updateNoteEditor){
	note=n;
	if (!note.isEmpty() && !systemFlags.isActive ("system-note"))
		systemFlags.activate ("system-note");
	if (note.isEmpty() && systemFlags.isActive ("system-note"))
		systemFlags.deactivate ("system-note");
}

QString TreeItem::getNote()
{
	return note.getNote();
}

bool TreeItem::hasEmptyNote()
{
	return note.isEmpty();
}

NoteObj TreeItem::getNoteObj()
{
	return note;
}

QString TreeItem::getNoteASCII(const QString &indent, const int &width)
{
    return note.getNoteASCII(indent,width);
}

QString TreeItem::getNoteASCII()
{
    return note.getNoteASCII();
}

QString TreeItem::getNoteOpenDoc()
{
    return note.getNoteOpenDoc();
}

void TreeItem::activateStandardFlag (const QString &name)
{
	standardFlags.activate (name);
	model->emitDataHasChanged(this);
}

void TreeItem::deactivateStandardFlag (const QString &name)
{
	standardFlags.deactivate (name);
	model->emitDataHasChanged(this);
}

void TreeItem::deactivateAllStandardFlags ()
{
	standardFlags.deactivateAll ();
	model->emitDataHasChanged(this);
}

void TreeItem::toggleStandardFlag(const QString &name, FlagRow *master)
{
	standardFlags.toggle (name,master);
	model->emitDataHasChanged(this);
}

bool TreeItem::isActiveStandardFlag (const QString &name)
{
	return standardFlags.isActive (name);
}

QStringList TreeItem::activeStandardFlagNames ()
{
	return standardFlags.activeFlagNames();
}

FlagRow* TreeItem::getStandardFlagRow()
{
	return &standardFlags;
}

/* FIXME-3 void TreeItem::updateToolBar()
{
	standardFlags.updateToolBar();
}
*/
QStringList TreeItem::activeSystemFlagNames ()
{
	return systemFlags.activeFlagNames();
}

bool TreeItem::canMoveDown()
{
	switch (type)
	{
		case Undefined: return false;
		case MapCenter: 
		case Branch: 
			if (!parentItem) return false;
			if (parentItem->num (this) < parentItem->branchCount()-1)
				return true;
			else
				return false;
			break;	
		case Image: return false;
		default: return false;
	}
}

bool TreeItem::canMoveUp()
{
	switch (type)
	{
		case Undefined: return false;
		case MapCenter: 
		case Branch: 
			if (!parentItem) return false;
			if (parentItem->num (this) > 0)
				return true;
			else
				return false;
			break;	
		case Image: return false;
		default: return false;
	}
}

void TreeItem::setID (const QString &s)
{
	objID=s;
}

QString TreeItem::getID()
{
	return objID;
}


TreeItem* TreeItem::getChildNum(const int &n)
{
	if (n>=0 && n<childItems.count() )
		return childItems.at(n);
	else
		return NULL;
}

BranchItem* TreeItem::getFirstBranch()
{
	if (branchCounter>0)
		return getBranchNum (branchOffset);
	else
		return NULL;
}

BranchItem* TreeItem::getLastBranch()
{
	if (branchCounter>0)
		return getBranchNum (branchOffset + branchCounter-1);
	else
		return NULL;
}

BranchItem* TreeItem::getNextBranch(BranchItem *currentBranch)
{
	if (!currentBranch) return NULL;
	int n=num (currentBranch)+1;
	if (n<branchCounter)
		return getBranchNum (branchOffset + n);
	else
		return NULL;
}


BranchItem* TreeItem::getBranchNum(const int &n)
{
	if (n>=0 && n<branchCounter)
		return (BranchItem*)getChildNum (branchOffset + n);
	else
		return NULL;
}

BranchObj* TreeItem::getBranchObjNum(const int &n)
{
	if (n>=0 && n<branchCounter)
	{
		return (BranchObj*)(getBranchNum(n)->getLMO());
	} else
		return NULL;
}

void TreeItem::setLastSelectedBranch()
{
	if (parentItem)
		parentItem->lastSelectedBranchNum=parentItem->childItems.indexOf(this);
}

void TreeItem::setLastSelectedBranch(int i)
{
		lastSelectedBranchNum=i;
}

TreeItem* TreeItem::getLastSelectedBranch()
{
	return getBranchNum (lastSelectedBranchNum);
}

ImageItem* TreeItem::getImageNum (const int &n)
{
	if (n>=0 && n<imageCounter)
		return (ImageItem*)getChildNum (imageOffset + n);
	else
		return NULL;
}

FloatImageObj* TreeItem::getImageObjNum (const int &n)	// FIXME-5 what about SVGs later?
{
	if (imageCounter>0 )
		return (FloatImageObj*)(getImageNum(n)->getLMO());
	else
		return NULL;
}


void TreeItem::setHideTmp (HideTmpMode mode)  //FIXME-2
{
	if (isBranchLikeType() )
		((BranchItem*)this)->updateVisibility();
		/*
	if (type==Image)
		//updateVisibility();
	*/

/*

		if (mode==HideExport && (hideExport || hasHiddenExportParent() ) )
		{
			// Hide stuff according to hideExport flag and parents
			//setVisibility (false);
			updateVisibility();
			//FIXME-2 hidden=true;
		}else
		{
			// Do not hide, but still take care of scrolled status
			

			XXXXXXXX treeItem should be THIS

			move visible to TreeItem???

			BranchObj now has updateContents

			maybe also start "bool TreeItem::branchlikeType"



			if ( ((BranchItem*)treeItem)->hasScrolledParent((BranchItem*)treeItem))
				setVisibility (false);
			else
				setVisibility (true);
			//FIXME-2 hidden=false;
		}	

*/
		// And take care of my children
		for (int i=0; i<branchCount(); ++i)
			getBranchNum(i)->setHideTmp (mode);
}

bool TreeItem::hasHiddenExportParent()
{
	// Calls parents recursivly to
	// find out, if we or parents are temp. hidden

	if (hidden || hideExport) return true;

	if (parentItem) 
		return parentItem->hasHiddenExportParent();
	else
		return false;
}


void TreeItem::setHideInExport(bool b) 
{
	if (type==Branch || type==Image)
	{
		hideExport=b;
		if (b)
			systemFlags.activate("system-hideInExport");
		else	
			systemFlags.deactivate("system-hideInExport");
	}
}	

bool TreeItem::hideInExport()
{
	return hideExport;
}	

bool TreeItem::isHidden()
{
	return hidden;
}	

QString TreeItem::getGeneralAttr()
{
	QString s;
	if (hideExport)
		 s+=attribut("hideInExport","true");
	if (!url.isEmpty())
		s+=attribut ("url",url);
	if (!vymLink.isEmpty())
		s+=attribut ("vymLink",vymLink);
	return s;	
}


